/*
 * ImageSender.c - image sender of the LOSSLESS DIGITAL IMAGE BROWSER
 *
 *  Author:    Rainer Lienhart, University of Mannheim, Germany
 *             (lienhart@pi4.informatik.uni-mannheim.de) or
 *             ICSI - International Computer Science Institue
 *             (lienhart@icsi.berkeley.edu)
 */

/* Copyright Notice
 * ================
 * Copyright 1994 by Rainer Lienhart
 * 
 * Permission to use, copy, and distribute LDIB in its entirety, for 
 * non-commercial purposes, is hereby granted without fee, provided that
 * this license information and copyright notice appear in all copies.
 * 
 * Note that distributing LDIB 'bundled' in with ANY product is considered
 * to be a 'commercial purpose'.
 *
 * The software may be modified for your own purposes, but modified
 * versions may NOT be distributed without prior consent of the author.
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the author be held liable for any damages
 * arising from the use of this software.
 * 
 * If you would like to do something with LDIB that this copyright
 * prohibits (such as distributing it with a commercial product), please
 * contact the author (preferably via email).  Arrangements can
 * probably be worked out.
 *
 * The author may be contacted via:
 *    Mail:     Rainer Lienhart
 *              Universitaet Mannheim
 *              Praktische Informatik IV
 *              68131 Mannheim
 *              Germany
 *
 *    Phone:    phone  : +49 621 292 1407
 *              fax    : +49 621 292 5745
 *    EMail:    lienhart@pi4.informatik.uni-mannheim.de
 */
#include "ImageSender.h"

/* ----------------------------------------------------------------------- */
/* Signal handler for SIGALRM, raised by the intervall timer.              */
/* It first executes received, but queued control instructions from the    */
/* Control Panel, then sends the next image to the Image Viewer.           */
/* ----------------------------------------------------------------------- */
void Catch_SIGALRM ()
{
  static char temp[40];
  /* process latest control instructions!                           */
  /* they are all delayed except "SPEED". "SPEED" is always handled */
  /* immediately.                                                   */
  if (latecom.position != -1) {
    actImageNumber = latecom.position % numberOfImages;
    latecom.position = -1;
    sprintf(temp,"SENT %d\n", actImageNumber);
    tested_write(cp,temp,strlen(temp));
    if ( latecom.speed == 0 ) {
      sendImage(actImageNumber);
    }
  }
  if (latecom.direction != -1) {
    direction = latecom.direction;
    latecom.direction = -1; 
    /* the direction change happens during the previous image !! */
    switch (direction) {
    case FORWARD: actImageNumber = (actImageNumber + 1) % numberOfImages;
      break;
    case BACKWARD: actImageNumber = ( actImageNumber ? (actImageNumber-1) : (numberOfImages-1) );
      break; 
    default: FatalError("unexpected");
      break;
    }
  }
  if (latecom.kill != -1) {
    tested_write(iv,"9999KILLKILL00000000",20);
    Quit(0);
  }

  if ( latecom.speed == 0 ) 
    return;

  /* Send next image */
  sendImage(actImageNumber);
  /* send number of currently sent image to the ControlPanel */ 
  sprintf(temp,"SENT %d\n", actImageNumber);
  tested_write(cp,temp,strlen(temp));

  /* set the number of the next image which is expected to be sent */
  switch (direction) {
  case FORWARD: actImageNumber = (actImageNumber + 1) % numberOfImages;
    break;
  case BACKWARD: actImageNumber = ( actImageNumber ? (actImageNumber-1) : (numberOfImages-1) );
    break; 
  default: FatalError("unexpected");
    break;
  }
}


/* ----------------------------------------------------------------------- */
/* Signal handler for SIGIO, raised by some data arriving from the         */
/* Control Panel. 
/* ----------------------------------------------------------------------- */
void Catch_SIGIO ()
{
  static char buffer[40], command[20], args[20];
  struct itimerval newTimer;
  int i,n, length;
  float m;

  /* test if data is available at the non-blocking socket from CP */
  while ( (i= read(cp,buffer,4)) > 0 ) { 
    sscanf(buffer,"%3d",&length);
    if ( (i = read(cp,buffer,length+1)) != length+1) {
      FatalError("Couldn't parse command from Control Panel\n");
    }
    buffer[i] = 0;

    /* process control instructions */
    sscanf(buffer,"%s %s",command,args);
    /*  -  -  -  -  -  -  -  SPEED  -  -  -  -  -  -  -  -   */
    if ( ! strncmp("SPEED",command,5) ) {
      sscanf(args,"%d",&latecom.speed);
      if (latecom.speed) {
	m = 10.0 / (float)latecom.speed;
	speed.tv_sec = (int) m;
	speed.tv_usec = (int) (m * 1000000) % 1000000;
      } else {
	latecom.speed = 0;
	speed.tv_sec = speed.tv_usec  = 0;
      }
      /* calculate parameter using policy1, 2 or 3 */
      /* e.g. latecom.policy(speed, numberOfImages, maxImageSize, ...) */
      /* issue  no-violation modification contract */
      /* guess that the following must decoupled */
      /* if modification contract has been successful */
      if (latecom.speed > 0) {
	sprintf(buffer,"SPEED %2d\n", latecom.speed);
	tested_write(cp,buffer,strlen(buffer));
      }
      /* set timer */
      newTimer.it_value.tv_usec = newTimer.it_interval.tv_usec = speed.tv_usec ;
      newTimer.it_value.tv_sec = newTimer.it_interval.tv_sec = speed.tv_sec;
      setitimer(ITIMER_REAL,&newTimer);

    /*  -  -  -  -  -  -  -  POSITION  -  -  -  -  -  -  -  -   */
    } else if ( ! strncmp("POSITION",command,strlen("POSITION")) ) {
      sscanf(args,"%d",&n);
      latecom.position = n % numberOfImages;
      if ( latecom.speed == 0) {
	/* set on interrupt, so the position is handled */
	newTimer.it_value.tv_usec = 100000 ;
	newTimer.it_interval.tv_usec = 0;
	newTimer.it_value.tv_sec = newTimer.it_interval.tv_sec = 0;
	setitimer(ITIMER_REAL,&newTimer);
      }
    /*  -  -  -  -  -  -  -  DIRECTION  -  -  -  -  -  -  -  -   */
    } else if ( ! strncmp("DIRECTION",command,strlen("DIRECTION")) ) {
      if ( !strncmp("BACKWARD",args,strlen("BACKWARD")) ) {
	if (direction == FORWARD) 
	  latecom.direction = BACKWARD;
      } else {
	latecom.direction = FORWARD;
      }
    /*  -  -  -  -  -  -  -  KILL  -  -  -  -  -  -  -  -   */
    } else if ( ! strncmp("KILL",command,strlen("KILL")) ) {
      latecom.kill = 1;
      newTimer.it_value.tv_usec = 10000 ;
      newTimer.it_interval.tv_usec = 0;
      newTimer.it_value.tv_sec = newTimer.it_interval.tv_sec = 0;
      setitimer(ITIMER_REAL,&newTimer);
    /*  -  -  -  -  -  -  -  POLICY  -  -  -  -  -  -  -  -   */
    } else if ( ! strncmp("POLICY",command,strlen("POLICY")) ) {
      sscanf(args,"%d",&n);
      switch (n) {
      case 1 : latecom.policy = policy1;
	break;
      case 2 : latecom.policy = policy2;
	break;
      case 3 : latecom.policy = policy3;
	break;
      default: latecom.policy = policy1;
      }
      /* changes only adaption policy; has only effect when  */
      /* speed is changed */
    }
  }
  buffer[0] = 0; 
}

/* ----------------------------------------------------------------------- */
/* setup a server port for connection between Image Sender and Control     */
/* Panel.                                                                  */
/* ----------------------------------------------------------------------- */
/* be careful : this function is blocking till the client
   connect to the socket 
*/
void setupServer(int port)
{
  struct hostent *hp;
  static struct sockaddr_in sa, in_sa; 
  char localhost [MAXHOSTNAME+1];
  int i;

  if ( gethostname (localhost, MAXHOSTNAME) == -1 ) 
    FatalError("gethostname");

  if ( (hp = gethostbyname(localhost)) == NULL)
    FatalError("gethostbyname");


  s = socket ( hp->h_addrtype, SOCK_STREAM, TCP);
  if (s == -1) FatalError("socket");

  /* struct sockaddr_in {
        short   sin_family;
        u_short sin_port;
        struct  in_addr sin_addr;
        char    sin_zero[8];
     };
  */
  bcopy((char *) hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;
  
  if (port == 0) {
    port = 5000;
    do {
      sa.sin_port = htons(port);
      
      if (bind(s, &sa, sizeof sa) > -1) {
	break;
      }
      port++;
    } while (port<20000);
    if (port == 20000) 
      FatalError("unable to bind ImageViewer");
  } else {
    sa.sin_port = htons(port); 
    if (bind(s, &sa, sizeof sa) == -1) 
      FatalError("unable to bind ImageViewer"); 
  }
  if (listen(s,1) == -1) FatalError("listen");

  /* write out serverportnumber for Tcl/Tk */
  printf("%d\n",port); fflush(stdout);

  i = sizeof in_sa;
  if ((cp = accept(s, &in_sa, &i)) < 0) FatalError("accept");

  fcntl(cp,F_SETOWN,getpid());
  /* must be set with one call ! */
  fcntl(cp,F_SETFL,FNDELAY | FASYNC); 

  signal(SIGIO,Catch_SIGIO);
  signal(SIGURG,Catch_SIGIO);
}

/* ----------------------------------------------------------------------- */
/* setup the connection to the Image Viewer.                               */
/* ----------------------------------------------------------------------- */
void setup_connection_to_ImageViewer(char *remoteHostName, int port)
{
  struct hostent *hp;
  static struct sockaddr_in sa, in_sa; 

  if ((hp = gethostbyname(remoteHostName)) == NULL) {
    fprintf(stderr, "ERROR : No host like '%s'\n", remoteHostName);
    exit(1);
  }
  /*
   * put host's address and address type into socket structure
   */
  bcopy((char *) hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;
  sa.sin_port = htons(port);
  if ((iv = socket(hp->h_addrtype, SOCK_STREAM, TCP)) < 0) {
    perror("socket");
    exit(1);
  }

   /* connect to the remote server */
  if (connect(iv,&sa, sizeof sa) < 0) FatalError("connect");
}


/* ----------------------------------------------------------------------- */
/* read the image (specified by number) from harddisk and send it to the   */
/* Image Viewer. It adds a header to each image.                           */
/* ----------------------------------------------------------------------- */
int sendImage(int number)
{
  static char buffer[20000];
  static char buffer2[10];
  FILE *fp;
  int i,n;

  fp = fopen(ImageList[number].ImageName,"r");
  if (!fp) {
    sprintf(buffer,"Can't open image %s",ImageList[number].ImageName);
    FatalError(buffer);
  }
  sprintf(buffer,"%04d",actImageNumber);
  tested_write(iv,buffer,4);
  
  fread(buffer,8,1,fp);
  tested_write(iv,buffer,8);
  sprintf(buffer2,"%08d",ImageList[number].ImageSize);
  tested_write(iv,buffer2,8);
  tested_write(iv,buffer,8);
  while (!feof(fp)) {
    i = fread(buffer,1,20000,fp);
    /* write all the data, but be aware that because of an 
       signal not all is writen at one time */
    n = 0;
    while ( n != i) {
      n+=tested_write(iv, buffer+n, i-n);
    }
  }
  fclose(fp);
}

/* ----------------------------------------------------------------------- */
/* identical to the system call write(), with the exception that it        */
/* terminates if an error occurs                                           */
/* ----------------------------------------------------------------------- */
int tested_write(int fd, char *buf, int nbyte)
{
  static int result;
  
  result = write(fd, buf, nbyte);
  if ( result < 0) {
    if (fd == iv)
      FatalError("lost connection to Image Viewer");
    else
      FatalError("lost connection to Control Panel");
  }
  return result;
}

/* ----------------------------------------------------------------------- */
/* identical to the system call read, with the exception that it           */
/* terminates if an error occurs                                           */
/* ----------------------------------------------------------------------- */
int tested_read(int fd, char *buf, int nbyte)
{
  static int result;
  
  result = read(fd, buf, nbyte);
  if ( result < 0) {
    if (fd == iv)
      FatalError("lost connection to Image Viewer");
    else
      FatalError("lost connection to Control Panel");
  }
  return result;
}

/* *********************************************************************** */
/* Main program of Image Sender                                            */
/* *********************************************************************** */
int main(int argc, char **argv)
{
  int i, imageViewerPort = -1, imageSenderPort = 0;
  char *imageViewer = NULL;
 
  /* parse command line */
  cmd = argv[0];
  for ( i=1 ; i<argc ; i++) {
    if ( (!strcmp(argv[i],"-imagelist")) || (!strcmp(argv[i],"-il")) ) {
      struct stat buf;
      ImageEntry *tmp;

      ImageList = tmp = (ImageEntry*) malloc ( sizeof(ImageEntry) * (argc - i) );
      i++;
      while ((i<argc) && (argv[i][0] != '-')) {
	tmp->ImageName = argv[i];
	if ( stat(argv[i],&buf) ) {
	  fprintf(stderr,"ERROR: '%s' is not a file\n",argv[i]);
	}
	tmp->ImageSize = buf.st_size;
        tmp++; i++;
	numberOfImages++;
      }
      i--;
      continue;
    }
    if ( (!strcmp(argv[i],"-imageviewer")) || (!strcmp(argv[i],"-iv")) ) {
      i++;
      imageViewer = argv[i];
      continue;
    }
    if ( (!strcmp(argv[i],"-imageviewerport")) || (!strcmp(argv[i],"-ivp")) ) {
      i++;
      sscanf(argv[i],"%d",&imageViewerPort);
      continue;
    }
    if ( (!strcmp(argv[i],"-imagesenderport")) || (!strcmp(argv[i],"-isp")) ) {
      i++;
      sscanf(argv[i],"%d",&imageSenderPort);
      continue;
    }
    if (!strncmp(argv[i],"-h",2)) {         /* help */
      Syntax();
    }
  }
  if ((imageViewerPort == -1) || (imageViewer == NULL) ||
      (numberOfImages == 0) )
    Syntax();

  maxImageSize = 0;
  /* determine size of largest image */
  for (i=0; i<numberOfImages; i++) {
    if (maxImageSize < ImageList[i].ImageSize)
      maxImageSize = ImageList[i].ImageSize;
  }

  signal(SIGINT,Catch_SIGTERM);
  signal(SIGTERM,Catch_SIGTERM);
  signal(SIGQUIT,Catch_SIGTERM);
  signal(SIGTSTP,Catch_SIGTERM);

  setup_connection_to_ImageViewer(imageViewer,imageViewerPort);
  setupServer(imageSenderPort);
  /* send number of images to the ControlPanel */ 
  sprintf(Text,"%d",numberOfImages);
  tested_write(cp,Text,strlen(Text));

  signal(SIGALRM,Catch_SIGALRM);
  while (1) 
    sigpause(0);
}

/* ----------------------------------------------------------------------- */
/* terminate program with an Error                                         */
/* ----------------------------------------------------------------------- */
void FatalError (char *text)
{
  fprintf(stderr, "ERROR in %s: %s\n",cmd, text);
  Quit(1);
}

/* ----------------------------------------------------------------------- */
/* print the SYNOPSIS to stdout                                            *
/* ----------------------------------------------------------------------- */
void Syntax()
{
  printf("SYNOPSIS: %s ",cmd);
  printf("-imagelist <pathname-list> "); 
  printf("-imageviewer <host address> "); 
  printf("-imageviewerport <int>"); 
  printf("[-imagesenderport <int>]\n"); 
  printf("SYNOPSIS: %s ",cmd);
  printf("-il <pathname-list> "); 
  printf("-iv <host address> "); 
  printf("-ivp <int>"); 
  printf("[-isp <int>]\n"); 
  printf("ALL PARAMETERS ARE MANDATORY!\n"); 
  exit(1);
}

/* ----------------------------------------------------------------------- */
/* First, cleanup everything, then terminate the program.                  */
/* ----------------------------------------------------------------------- */
void Quit(int exitcode)
{
  fprintf(stderr, "Terminating %s\n", cmd);
  if (iv) close(iv); 
  if (cp) close(cp);
  if (s)  close(s);
  free(ImageList);
  exit(exitcode);
}

/* ----------------------------------------------------------------------- */
/* Signal handler for all signals which cause the program to be terminated */
/* ----------------------------------------------------------------------- */
void Catch_SIGTERM ()
{
  Quit(0);
}

/* ----------------------------------------------------------------------- */
/* implementation of policy 1:                                             */
/* immediate adaption/adjustment of the traffic & performance parameters   */
/* ----------------------------------------------------------------------- */
void policy1(int speed, int numberOfImages, long maxImageSize,
	     rcapTraffic *traffic, rcapRequirements *requirements)
{
  long maxsize = 0;
  int i;

  /* standard temporal units */
  traffic->xmin = (u_int)( (10.0 / (float)speed) * (float)0x10000);
  traffic->xave = traffic->xmin;
  traffic->I = traffic->xmin * numberOfImages;
  traffic->smax = maxImageSize + 20;
  
  /* standard temporal units */
  requirements->D = 0x1999;
  requirements->J = (int)((float)(traffic->xave) * .05);
  /* standard probability units */
  requirements->Z = 10000;
  requirements->W = 0;
  requirements->U = 0;
}

/* ----------------------------------------------------------------------- */
/* implementation of policy 2:                                             */
/* immediate adaption/adjustment of the traffic & performance parameters   */
/* with least resource reservation                                         */
/* ----------------------------------------------------------------------- */
void policy2(int speed, int numberOfImages, long maxImageSize,
	     rcapTraffic *traffic, rcapRequirements *requirements)
{
  int i;
  u_int tempxave, minxave = 0x20000;

  /* standard temporal units */
  tempxave = (u_int)( (10.0 / (float)speed) * (float)0x10000);
  if (tempxave < minxave)
    tempxave = minxave;
  traffic->xmin = tempxave;
  traffic->xave = traffic->xmin;
  traffic->I = traffic->xmin * numberOfImages;
  traffic->smax = maxImageSize + 20;
  
  /* standard temporal units */
  requirements->D = 0x1999;
  requirements->J = (int)((float)(traffic->xave) * .05);
  /* standard probability units */
  requirements->Z = 10000;
  requirements->W = 0;
  requirements->U = 0;
}

/* ----------------------------------------------------------------------- */
/* implementation of policy 3:                                             */
/* smooth adaption/adjustment of the traffic & performance parameters      */
/* ----------------------------------------------------------------------- */
void policy3(int speed, int numberOfImages, long maxImageSize,
	     rcapTraffic *traffic, rcapRequirements *requirements)
{
  /* not implemented yet */
}
