/*
 * ImageViewer.c - image viewer of the LOSSLESS DIGITAL IMAGE BROWSER
 *
 *  Author:    Rainer Lienhart, University of Mannheim, Germany
 *             (lienhart@pi4.informatik.uni-mannheim.de) or
 *             ICSI - International Computer Science Institue
 *             (lienhart@icsi.berkeley.edu)
 */

/* Copyright Notice
 * ================
 * Copyright 1994 by Rainer Lienhart
 * 
 * Permission to use, copy, and distribute LDIB in its entirety, for 
 * non-commercial purposes, is hereby granted without fee, provided that
 * this license information and copyright notice appear in all copies.
 * 
 * Note that distributing LDIB 'bundled' in with ANY product is considered
 * to be a 'commercial purpose'.
 *
 * The software may be modified for your own purposes, but modified
 * versions may NOT be distributed without prior consent of the author.
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the author be held liable for any damages
 * arising from the use of this software.
 * 
 * If you would like to do something with LDIB that this copyright
 * prohibits (such as distributing it with a commercial product), please
 * contact the author (preferably via email).  Arrangements can
 * probably be worked out.
 *
 * The author may be contacted via:
 *    Mail:     Rainer Lienhart
 *              Universitaet Mannheim
 *              Praktische Informatik IV
 *              68131 Mannheim
 *              Germany
 *
 *    Phone:    phone  : +49 621 292 1407
 *              fax    : +49 621 292 5745
 *    EMail:    lienhart@pi4.informatik.uni-mannheim.de
 */
#include "ImageViewer.h"

/* ----------------------------------------------------------------------- */
/* calculates current and average frame per second; sends it to the        */
/* Control Panel                                                           */
/* ----------------------------------------------------------------------- */
void calcAndSendFPS() {
  static struct timeb newTime;   /* time of arrival of current image */
  static struct timeb startTime; /* time of first arrived image */
  static struct timeb oldTime = {0,0,0,0}; /* time of arrival of previous image */
  static float intervall, fps;
  static int numOfViewedImages = 0;
  static char bufffer[100];

  /* first call is used to initialize all variables */
  if (oldTime.time == 0) {
    ftime(&oldTime);
    ftime(&startTime);
    return;
  }
      
  ftime(&newTime); 
  numOfViewedImages++;
  /* calculate FPS & send to Control Panel */
  intervall = (float)(newTime.time - oldTime.time); 
  intervall += (float)(newTime.millitm - oldTime.millitm) / (float) 1000; 
  fps = 1/intervall; 
  sprintf(buffer,"FPS %f\n", fps);
  tested_write(cp,buffer,strlen(buffer));

  /* calculate AVE_FPS & send to Control Panel */
  intervall = (float)(newTime.time - startTime.time); 
  intervall += (float)(newTime.millitm - startTime.millitm) / (float) 1000000;
  intervall /= (float)numOfViewedImages;
  fps = 1/intervall; 
  sprintf(buffer,"AVE_FPS %f\n", fps);
  tested_write(cp,buffer,strlen(buffer));

  /* set new oldTime */
  oldTime.time = newTime.time ; 
  oldTime.millitm = newTime.millitm; 
}

/* ----------------------------------------------------------------------- */
/* setup the Server port for the Real-time Channel (RTC)                   */
/* be careful : this function is blocking till the client                  */
/* connects to the socket                                                  */
/* ----------------------------------------------------------------------- */
int setupServerRTC(int port)
{
  struct hostent *hp;
  static struct sockaddr_in sa, in_sa; 
  char localhost [MAXHOSTNAME+1];
  int i;

  if ( gethostname (localhost, MAXHOSTNAME) == -1 ) { 
     FatalError("setupServer","while executing 'gethostname'");
  }

  if ( (hp = gethostbyname(localhost)) == NULL) {
    FatalError("setupServer","while executing 'gethostbyname'");
  }

  s = socket ( hp->h_addrtype, SOCK_STREAM, TCP);
  if (s == -1) { FatalError("setupServer","while executing 'socket'"); }

  /* struct sockaddr_in {
        short   sin_family;
        u_short sin_port;
        struct  in_addr sin_addr;
        char    sin_zero[8];
     };
  */
  bcopy( (char *)hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;

  /* search for free portnumber, starting with 5000 */
  if (port == 0) {
    port = 5000;
    do {
      sa.sin_port = htons(port);
      
      if (bind(s, &sa, sizeof sa) > -1) {
	break;
      }
      port++;
    } while (port<10000);
    if (port == 10000)
      FatalError("setupServer","unable to bind ImageViewer");
  } else {
    sa.sin_port = htons(port); 
    if (bind(s, &sa, sizeof sa) == -1) 
      FatalError("setupServer","RTCbind");
  }

  if (listen(s,1)  == -1) { 
    FatalError("setupServer","while executing 'listn'");
  }

    printf("%d\n",port);
    fflush(stdout);

  return port;
}

/* ----------------------------------------------------------------------- */
/* setup up the TCP channel to the Control Panel; used to receive the      */
/* control instruction                                                     */
/* be careful : this function is blocking till the Control Pannel and      */
/* Image Viewer connects to the socket                                     */
/* ----------------------------------------------------------------------- */
int setupServerTCP(int port)
{
  struct hostent *hp;
  static struct sockaddr_in sa, in_sa; 
  char localhost [MAXHOSTNAME+1];
  int i;

  if ( gethostname (localhost, MAXHOSTNAME) == -1 ) { 
     FatalError("setupServer","while executing 'gethostname'");
  }

  if ( (hp = gethostbyname(localhost)) == NULL) {
    FatalError("setupServer","while executing 'gethostbyname'");
  }

  s2 = socket ( hp->h_addrtype, SOCK_STREAM, TCP);
  if (s2 == -1) { FatalError("setupServerTCP","while executing 'socket'"); }

  /* struct sockaddr_in {
        short   sin_family;
        u_short sin_port;
        struct  in_addr sin_addr;
        char    sin_zero[8];
     };
  */
  bcopy( (char *)hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  sa.sin_family = hp->h_addrtype;

  if (port == 0) {
    /* search for free portnumber, starting with 5000 */
    port = 5000;
    do {
      sa.sin_port = htons(port);
      
      if (bind(s2, &sa, sizeof sa) > -1) {
	break;
      }
      port++;
    } while (port<10000);
    if (port == 10000) {
      FatalError("setupServerTCP","unable to bind ImageViewer");
    }
  } else {
    sa.sin_port = htons(port); 
    if (bind(s2, &sa, sizeof sa) == -1) 
      FatalError("setupServerTCP","TCPbind");
  } 

  if (listen(s2,1)  == -1) { 
    FatalError("setupServer","while executing 'listn'");
  }
  
  printf("%d\n",port);
  fflush(stdout);

  i = sizeof in_sa;
  if ((is = accept(s, &in_sa, &i)) < 0) {
    FatalError("setupServerTCP","while executing 'accept'");
  }
  i = sizeof in_sa;
  if ((cp = accept(s2, &in_sa, &i)) < 0) {
    FatalError("setupServerTCP","while executing 'accept'");
  }
  return port;
}


/* *********************************************************************** */
/* Main program of Image Viewer                                            */
/* *********************************************************************** */
main(int argc, char* argv[])
{
    int       i;
    char      *display, *geom;
    XEvent    event;
    int       realtimePort, tcpPort;

    cmd = argv[0];
    display = geom = NULL;
    realtimePort = tcpPort = 0;

    /*============parse command options============*/
    for (i = 1; i < argc; i++) {
      char *strind;
      
      if (!strncmp(argv[i],"-d",2)) {	      /* display */
	i++;
	display = argv[i];
	continue;
      }
      if (!strncmp(argv[i],"-g",2)) {	      /* geometry */
	i++;
	geom = argv[i];
	continue;
      }
      if ((!strncmp(argv[i],"-rtcp",2)) || 
	  (!strncmp(argv[i],"-realtimechannelport",15)) ) {
	i++;
	sscanf(argv[i],"%d",&realtimePort);
	continue;
      }
      if ((!strncmp(argv[i],"-tcpp",2)) || 
	  (!strncmp(argv[i],"-tcpport",15)) ) {
	i++;
	sscanf(argv[i],"%d",&tcpPort);
	continue;
      }

      if (!strncmp(argv[i],"-h",2)) {         /* help */
	Syntax();
      }
    }

    signal(SIGINT,Catch_SIGTERM);
    signal(SIGTERM,Catch_SIGTERM);
    signal(SIGQUIT,Catch_SIGTERM);
    signal(SIGTSTP,Catch_SIGTERM);
    signal(SIGPIPE,Catch_SIGTERM);

    /* install server ports */
    realtimePort = setupServerRTC(realtimePort);
    tcpPort  = setupServerTCP(tcpPort); 
    while ( ImageDataAtSocketToRead (cp) < 0) ;
    tested_read(cp,buffer,100);
    sscanf(buffer,"%*s%*2s%x",&CP_XID);
    signal(SIGPIPE,SIG_IGN);


    /* Open up the display. */
    if ( (theDisp=XOpenDisplay(display)) == NULL)
        FatalError("main","can't open display");

    theScreen = DefaultScreen(theDisp);
    theCmap   = DefaultColormap(theDisp, theScreen);
    rootW     = RootWindow(theDisp,theScreen);
    theGC     = DefaultGC(theDisp,theScreen);
    fcol      = WhitePixel(theDisp,theScreen);
    bcol      = BlackPixel(theDisp,theScreen);
    theVisual = DefaultVisual(theDisp,theScreen);
    theDepth  = DefaultDepth(theDisp,theScreen);

    /**************** Create/Open X Resources ***************/
    if ((mfinfo = XLoadQueryFont(theDisp,FONT))==NULL) {
       sprintf(tmpstr,"couldn't open '%s' font",FONT);
       FatalError("main",tmpstr);
       }

    mfont=mfinfo->fid;
    XSetFont(theDisp,theGC,mfont);
    XSetForeground(theDisp,theGC,fcol);
    XSetBackground(theDisp,theGC,bcol);

    CreateMainWindow(cmd,geom,argc,argv);

    /* Lese das erste Bild ein */
    LoadPic();
    Resize((*test).w,(*test).h);

    XSelectInput(theDisp, mainW, ExposureMask | KeyPressMask | ColormapNotify
                               | StructureNotifyMask | ButtonPressMask);
    XMapWindow(theDisp,mainW);

    i = 2;
    /**************** Main loop *****************/
   
    while (1) {
      while ( XCheckMaskEvent(theDisp, 0xFFFFFF, &event) ) 
	HandleEvent(&event);
      if (xi) { 
	if ( ImageDataAtSocketToRead (is) > 0) {
	  LoadPic(); 
	  i++;
	}
      }
    }
  }

/* ----------------------------------------------------------------------- */
/* tests if date is currently present to read from socket t                */
/*     return 0 = no data to read                                          */
/*     return 1 = data to read                                             */
/* ----------------------------------------------------------------------- */
int ImageDataAtSocketToRead (int t) 
{
  fd_set listOfFileDescriptors;
  struct timeval TimeOut;

  TimeOut.tv_sec = 0 ; TimeOut.tv_usec = 10000;
  FD_ZERO(&listOfFileDescriptors);
  FD_SET(is,&listOfFileDescriptors);
  return select(ulimit(4),&listOfFileDescriptors,NULL,NULL,&TimeOut);
}
  
/* ----------------------------------------------------------------------- */
/* identical to the system call write(), with the exception that it        */
/* terminates if an error occurs                                           */
/* ----------------------------------------------------------------------- */
int tested_write(int fd, char *buf, int nbyte)
{
  static int result;
  
  result = write(fd, buf, nbyte);
  if ( result < 0) {
    if (fd = cp)
      FatalError("tested_write","lost connection to Control Panel");
    else
      FatalError("tested_read","fatal error");
  }
  return result;
}

/* ----------------------------------------------------------------------- */
/* identical to the system call read, with the exception that it           */
/* terminates if an error occurs                                           */
/* ----------------------------------------------------------------------- */
int tested_read(int fd, char *buf, int nbyte)
{
  static int result;
  
  result = read(fd, buf, nbyte);
  if ( result < 0) {
    if (fd = is)
      FatalError("tested_read","lost connection to ImageViewer");
    else
      FatalError("tested_read","fatal error");
  }
  return result;
}


/* ----------------------------------------------------------------------- */
/* handle XEvent                                                           */
/* ----------------------------------------------------------------------- */
int HandleEvent(XEvent *event)
{
    switch (event->type) {
        case Expose: {
            XExposeEvent *exp_event = (XExposeEvent *) event;

	    /* skip all expose events but the last one */
	    XEvent temp;
	    do {
	      if (!XCheckMaskEvent(theDisp, 0xffffff, &temp))
		break;
	      if (temp.type == Expose) {
		bcopy ((char*)&temp,(char*)exp_event, sizeof(XExposeEvent));
	      } else {
		XPutBackEvent(theDisp, &temp);
	      }
	    } while (temp.type == Expose);


            if ( (exp_event->window==mainW) && (exp_event->count == 0) ) 
	      if (!xi) {
		xi=XCreateImage(theDisp, theVisual,8,ZPixmap,0,(char*)(*test).pic,
				(*test).w,(*test).h,32,(*test).w);
		XPutImage(theDisp, mainW, theGC, xi ,0,0,0,0,(*test).w,(*test).h); 
	      } else {
		XPutImage(theDisp, mainW, theGC, xi ,0,0,0,0,(*test).w,(*test).h); 
	      } 
	  }
	  break;

/*        case ButtonPress: {
            XButtonEvent *but_event = (XButtonEvent *) event;
            if (but_event->window == mainW && but_event->button == Button1) ;
            }
            break;

        case KeyPress: {
            XKeyEvent *key_event = (XKeyEvent *) event;
            KeySym ks;
            XComposeStatus status;

            XLookupString(key_event,tmpstr,128,&ks,&status);
            if (tmpstr[0]=='q' || tmpstr[0]=='Q') Quit(0);
            }
            break;
*/
        case ConfigureNotify: {
            XConfigureEvent *conf_event = (XConfigureEvent *) event;
            if (conf_event->window == mainW && 
                 (conf_event->width != WIDE || conf_event->height != HIGH))
                Resize(conf_event->width, conf_event->height);
            }
            break;

	case ColormapNotify: 
        case CirculateNotify:
        case MapNotify:
        case DestroyNotify:
        case GravityNotify:
        case ReparentNotify:
        case UnmapNotify:
	case MappingNotify:      break;

        default:		/* ignore unexpected events */
	  break;
        }  /* end of switch */
}

/* ----------------------------------------------------------------------- */
/* prints command line options of the Image Viewer to stdout               */
/* ----------------------------------------------------------------------- */
void Syntax()
{
  printf("SYNOPSIS: %s",cmd);
  printf(" [-realtimechannelport <int>]"); 
  printf(" [-tcpport <int>]"); 
  printf(" [-geometry 'geometry']"); 
  printf(" [-display 'display']\n"); 
  printf("SYNOPSIS: %s ",cmd);
  printf(" [-rtcp <int>]"); 
  printf(" [-tcpp <int>]"); 
  printf(" [-g 'geometry']"); 
  printf(" [-d 'display']\n"); 
  exit(1);
}

/* ----------------------------------------------------------------------- */
/* terminate program with an Error                                         */
/* ----------------------------------------------------------------------- */
void FatalError (char *function, char *text)
{
  /* now, as the program is terminated, all signals should be ignored to */
  /* avoid infinite recursion (e.g. with SIGPIPE)                        */
  signal(SIGPIPE,SIG_IGN);
  signal(SIGINT,SIG_IGN);
  signal(SIGTERM,SIG_IGN);
  signal(SIGQUIT,SIG_IGN);
  signal(SIGTSTP,SIG_IGN);
  fprintf(stderr, "ERROR in %s: %s\n",cmd, text);
  Quit(1);
}

/* ----------------------------------------------------------------------- */
/* terminate program with the specified exit code                          */
/* ----------------------------------------------------------------------- */
void Quit(int exitcode)
{
  /* now, as the program is terminated, all signals should be ignored to */
  /* avoid infinite recursion (e.g. with SIGPIPE)                        */
  signal(SIGPIPE,SIG_IGN);
  signal(SIGINT,SIG_IGN);
  signal(SIGTERM,SIG_IGN);
  signal(SIGQUIT,SIG_IGN);
  signal(SIGTSTP,SIG_IGN);
  fprintf(stderr, "Terminating %s\n", cmd);
  if (cp) tested_write(cp,"KILL",4); /* does not arrive */
  if (is) close(is); 
  if (s) close(s); 
  if (cp) close(cp);
  if (s2)  close(s2);
  if (xi) XDestroyImage(xi);
 
  /* Colormap = XID = unsigned long */
  if (my_colormap) {
    XUninstallColormap(theDisp,my_colormap);
    XFreeColormap(theDisp,my_colormap);
  }
  exit(exitcode);
}

/* ----------------------------------------------------------------------- */
/* create the main window, that is the window the image are shown          */
/* ----------------------------------------------------------------------- */
void CreateMainWindow(char *name,char *geom, int argc,char **argv)
{
    XSetWindowAttributes xswa;
    unsigned int xswamask;
    XSizeHints hints;
    Pixmap icon;
    int i,x,y,w,h;

    WIDE = 640; HIGH = 400;    	 /* default window size */

    x=y=w=h=1;
    i=XParseGeometry(geom,&x,&y,&w,&h);
    if (i&XValue || i&YValue) hints.flags = USPosition;  
                         else hints.flags = PPosition;

    hints.flags |= PSize; /* application sets window size */ 

    if (i&XValue && i&XNegative) 
        x = XDisplayWidth(theDisp,theScreen)-WIDE-abs(x);
    if (i&YValue && i&YNegative) 
        y = XDisplayHeight(theDisp,theScreen)-HIGH-abs(y);

    hints.x=x;             hints.y=y;
    hints.width  = WIDE;   hints.height = HIGH;
    hints.max_width  = DisplayWidth(theDisp,theScreen);
    hints.max_height = DisplayHeight(theDisp,theScreen);
    hints.min_width  = 16;
    hints.min_height = 16;
    hints.width_inc = hints.height_inc = 16;
    hints.flags |= PMaxSize | PMinSize | PResizeInc;

    xswa.background_pixel = bcol;
    xswa.border_pixel     = fcol;
    xswa.cursor = XCreateFontCursor (theDisp, XC_top_left_arrow);
    xswamask = CWBackPixel | CWBorderPixel | CWCursor;

    mainW = XCreateWindow(theDisp,rootW,x,y,WIDE,HIGH,2,0,CopyFromParent,
                          CopyFromParent, xswamask, &xswa);

    XReadBitmapFile(theDisp, rootW, "/tmp_mnt/n/tenet/db/users/lienhart/LDIB/bitmaps/Logo.bit",&w,&h,&icon,&x,&y);

    XSetStandardProperties(theDisp,mainW,name,name,icon,
                            argv,argc,&hints);
    if (!mainW) FatalError("CreateMainWindow","Can't open main window");

}

/* ----------------------------------------------------------------------- */
/* change the size of the window; necassary if images of different size    */
/* are displayed                                                           */
/* ----------------------------------------------------------------------- */
void Resize(int w, int h)
{
  WIDE = w;  HIGH = h;
  XResizeWindow(theDisp, mainW, WIDE, HIGH);
}
                
/* ----------------------------------------------------------------------- */
/* loads the next image from the connection to the Image Sender, decodes   */
/* it, sets the colors and displays it                                     */ 
/* ----------------------------------------------------------------------- */
/* reads one image from socket */
int LoadPic()
{
  static int pic_nr = 0;            /* for alternating buffer */
  PICINFO *oldpicinfo;
  static char ApplicationName[STRINGLENGTH];
  int i,n, ii, imagetype;
  XVisualInfo visual_info;

  /* generate Window Title Name with current image number */
  tested_read(is,buffer,4);
  buffer[4] = 0;
  strncpy(ApplicationName,cmd,STRINGLENGTH - 7);
  strcat(ApplicationName,"  ");
  strncat(ApplicationName,buffer,4);
  XStoreName(theDisp,mainW,ApplicationName);

  test = &picinfo[pic_nr&1];
  oldpicinfo = &picinfo[(pic_nr+1)&1];

  /* unknown image format ? */
  imagetype = ReadImageType();
  if (imagetype == RFT_UNKNOWN) {
    FatalError ("LoadPic","unknown image type");
  }
  ReadImage(imagetype, test);
  changePalette((unsigned long*) test->pic, (test->w)*(test->h));

  /* execute only for the first image */
  if (pic_nr == 0) {
    if (XMatchVisualInfo(theDisp, theScreen, theDepth, PseudoColor, &visual_info))
      ; /*  printf("PseudoColor\n"); */
    else if (XMatchVisualInfo(theDisp, theScreen, theDepth, DirectColor, &visual_info))
      ; /* printf("DirectColor\n"); */
    else 
      FatalError("LoadPic","need PseudoColor or DirectColor Display");

    ncolors = DisplayCells(theDisp,theScreen);
    if (ncolors > 256) ncolors = 256;

    def_colormap = DefaultColormap(theDisp,theScreen);
    for (i=0; i<ncolors ; i++) {
      xcolor[i].pixel = i;
      xcolor[i].flags = DoRed|DoGreen|DoBlue;
    }
    XQueryColors(theDisp, def_colormap, xcolor, ncolors);
    my_colormap = XCreateColormap(theDisp,mainW,
				  theVisual,
				  AllocAll);
  }

  for (i=0, ii = 2; ii<ncolors ; i++, ii++) {
    xcolor[ii].red = (*test).r[i]<<8;
    xcolor[ii].blue = (*test).b[i]<<8;
    xcolor[ii].green = (*test).g[i]<<8;
    xcolor[ii].flags = DoRed|DoGreen|DoBlue;
    xcolor[ii].pixel = ii;
  } 

  /* Window is already mapped */
  if (pic_nr != 0) {
    /* it's not need because a global memory (spreicher) is used */
    /* free(oldpicinfo->pic); */
    if ((test->w == oldpicinfo->w) && 
	(test->h == oldpicinfo->h)) {
      xi->data = test->pic;
    } else {
      XDestroyImage(xi);
      xi=XCreateImage(theDisp, theVisual,8,ZPixmap,0,(char*)(*test).pic,
		      (*test).w,(*test).h,32,(*test).w);
      Resize((*test).w,(*test).h);
    }
    XPutImage(theDisp, mainW, theGC, xi ,0,0,0,0,test->w,test->h); 
  }

  XStoreColors(theDisp, my_colormap, xcolor, ncolors); 

  if (pic_nr == 0) {
    XSetWindowColormap(theDisp,mainW,my_colormap);
    XSetWindowColormap(theDisp,CP_XID,my_colormap);
    XInstallColormap(theDisp, my_colormap);
  }

  XFlush(theDisp);
  calcAndSendFPS();
  pic_nr++;
  return 0;
}

/* ----------------------------------------------------------------------- */
/* shifts all color values 2 entries higher                                */
/* ----------------------------------------------------------------------- */
int changePalette(unsigned long *data, int size)
{
  int i;
  size /= sizeof(unsigned long);
  for (i=0 ; i<size ; i++ ) {
    data[i] += 0x02020202;
  }
  return 0;
}
  
/* ----------------------------------------------------------------------- */
/* determine type of image present at the real-time channel                */
/* ----------------------------------------------------------------------- */
int ReadImageType()
{
  /* reads the first couple o' bytes.  Figures out what the file's likely
     to be, and returns the appropriate RFT_*** code */
  
  static byte  magicno[8];    /* first 8 bytes of stream */
  int rv = RFT_UNKNOWN;
  
  rv = tested_read(is,magicno,8);
  if (rv != 8){ 
    /* files less than 8 bytes long... */
    FatalError("ReadImageType","corrupted image data");     
  }
  
  if (strncmp((char *) magicno,"GIF87a",6)==0 ||
      strncmp((char *) magicno,"GIF89a",6)==0)
    rv = RFT_GIF;
  /* ==== insert new decoder recognitions here        === */
  /* === e.g.                                         === */
  /* === if (strncmp((char *) magicno,"IF_XXX",6)==0) === */
  /* ===   rv = RFT_XXX;                              === */
  else if (strncmp((char *) magicno,"KILLKILL",8)==0)
    Quit(0);
  
  return rv;
}

/* ----------------------------------------------------------------------- */
/* read one image of type 'imageType' from real-time channel into PICINFO  */
/* ----------------------------------------------------------------------- */
int ReadImage(int imageType, PICINFO *pinfo)
{
  int rv = 0;
  
  /* by default, most formats aren't multi-page */
  /* used because of using XV-loader */
  pinfo->numpages = 1;
  pinfo->pagebname[0] = '\0';
  
  switch (imageType) {
  case RFT_GIF:     rv = LoadGIF(" ", pinfo);  break;
  /* === add new case statements for new image formats here === */
  /* === e.g. "case RFT_XXX: rv = LOADXXX(pinfo);"          === */
  }
  return rv;
}

/* ----------------------------------------------------------------------- */
/* signal handler, used by all signals which should cause termination      */
/* ----------------------------------------------------------------------- */
void Catch_SIGTERM ()
{
  Quit(0);
}
