#ifndef RCAP_H
#define RCAP_H
/*
 * rcap.h
 * Bruce A. Mah (bmah@tenet.berkeley.edu)
 *
 * Exports RCAP-User interface to client programs.
 *
 * $Log: rcap.h,v $
 * Revision 2.0  1994/02/26  12:00:10  banerjea
 * Rcap daemon. Version 2.0
 *
 * Revision 1.6  1994/02/26  09:00:55  banerjea
 * New release contains the following features
 * Release resource bug fix
 * CM support
 * RMTP fragmentation support
 * Enhanced robustness i.e. any control socket being closed
 * causes all channels going through in to go away.
 * support of reset_rcap which removes all channels in a node
 *
 * Revision 1.1  1994/02/10  00:00:33  banerjea
 * Initial revision
 *
 * Revision 1.5  1993/06/11  01:08:15  banerjea
 * Changed the name of in_addr to avoid possible conflicts with Unix usage
 *
 * Revision 1.4  1993/06/05  00:27:22  banerjea
 * Removed the definitionof RECEIVE_BUFFER_SIZE because it
 * already eists in config.h
 *
 * Revision 1.3  1992/08/17  17:46:39  bmah
 * Added copyright stuff to keep the lawyers happy.
 *
 * Revision 1.2  1992/05/21  03:32:23  bmah
 * Misc fixes.
 *
 * Revision 1.1  1992/01/18  22:47:01  bmah
 * Initial revision
 *
 * Copyright (c) 1992 Regents of the University of California
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research, and non-profit purposes and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of the University of California not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Permission to incorporate this
 * software into commercial products can be obtained from the Campus
 * Software Office, 295 Evans Hall, University of California, Berkeley,
 * Ca., 94720 provided only that the the requestor give the University
 * of California a free licence to any derived software for educational
 * and research purposes.  The University of California makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 */

#include "consts.h"

/*
 * Type definitions for the RCAP design document.  These
 * may go away in favor of a plain "int", "short", and
 * "char".
 */
#include <sys/types.h>

/*
 * Other includes
 */
#include <netinet/in.h>

/*
 * Constants
 */

/*
 * Data types for RCAP structures.
 *
 * rcapTraffic is a structure which holds traffic characteristics.
 * This characterization is on a message (not packet) level.
 */
typedef struct {
   unsigned long xmin;			/* minimum interarrival time */
   unsigned long xave;			/* average interarrival time */
   unsigned long I;			/* averaging interval */
   unsigned long smax;			/* maximum message size in bytes */
} rcapTraffic;

/*
 * rcapRequirements is a structure holding performance requirements.
 * As for traffic characteristics, these figures are on a per-message
 * basis.
 */
typedef struct {
   unsigned long D;			/* end-to-end delay */
   unsigned long J;			/* delay jitter */
   unsigned long Z;			/* statistical delay probability */
   unsigned long W;			/* statistical no-drop probability */
   unsigned long U;			/* statistical jitter probability */
   unsigned char type;		/* channel type */
} rcapRequirements;

/*
 * in_addr holds an Internet-style address.
 */
typedef struct {
   unsigned long addr;
} rcap_in_addr;

/*
 * rcapAddress contains the IP address and port number of an entity in
 * the network.
 */
typedef struct {
    rcap_in_addr ipAddr;
   unsigned short port;
} rcapAddress;

/*
 * rcapUserControl contains information on the user control info sent
 * in an establishment request.
 */
typedef struct {
    void *userControl;
   unsigned short userControlLength;
} rcapUserControl;

/*
 * A parmblock is just an easy way of getting at the different
 * parameters of a call establishment.
 */
typedef struct {
    rcapTraffic *traffic;
    rcapRequirements *requirements;
    rcapAddress *destination;
    rcapUserControl *control;
} parmblock;

typedef struct BufferStruct
{
  unsigned long		BufferLength;
  unsigned long		NumDesc;
  unsigned long		ShmBuffer;
  unsigned long		ShmDesc;
  unsigned long		ShmSync;
  unsigned long		UserPID;
} bufferstruct;

typedef struct
{
  unsigned long     STDUmax;
  unsigned char     CONST_SIZE;
  unsigned char     CONST_NUM;
  unsigned long     T;
  unsigned long     Nmax;
  unsigned long     Smax;
  unsigned long     Savg;
  unsigned long     Smin;
  unsigned long     Iavg;
  unsigned long     Slack;
  bufferstruct *Buffer;
} rcapCmTraffic;

typedef struct
{
  unsigned long     Dstrm;
  unsigned long     Werr;
  unsigned long     Serr;
  unsigned char     Replace;
  unsigned char    *Dummy;
} rcapCmRequirements;

typedef struct CmParmBlock
{
  rcapCmTraffic		*CmTraffic;
  rcapCmRequirements 	*CmRequirements ;
  rcapAddress 		*destination;
  rcapUserControl	*control;
} cmParmblock;

/*
 * User-callable RCAP routines
 */
extern unsigned long RcapEstablishRequest(parmblock *parms, unsigned short *lcid, struct in_addr *ipAddr);
extern unsigned long RcapStatusRequest(unsigned short lcid, void *statusBuffer, unsigned short *statusBufferLength);

extern unsigned long RcapCloseRequest(unsigned short lcid, unsigned short reasonCode);

extern unsigned long RcapRegister(unsigned short port, unsigned long queue_length);
extern unsigned long RcapReceiveRequest(unsigned short port, parmblock *parms, rcapAddress
			 *source, unsigned short *lcid);
extern unsigned long RcapEstablishReturn(unsigned short lcid, unsigned short result, unsigned short
			  reasonCode, rcapUserControl *control);
extern unsigned long RcapUnregister(unsigned short port);

#define RcapCmCloseRequest RcapCloseRequest
extern unsigned long RcapCmEstablishRequest(cmParmblock *, unsigned short *, struct in_addr *);
extern unsigned long RcapCmReceiveRequest(unsigned short, cmParmblock *, rcapAddress *, unsigned short *);
extern unsigned long RcapCmEstablishReturn(unsigned short, unsigned short, unsigned short, cmParmblock *);

extern long CmBufferRequirement(rcapCmTraffic *traffic, unsigned long reqDstrm);

extern char* RcapStrError(unsigned long);

#endif RCAP_H

